@@ -1,7 +1,7 @@
 package Web::Async::WebSocket::Server::Connection;
 use Myriad::Class extends => 'IO::Async::Notifier';
 
-our $VERSION = '0.002'; ## VERSION
+our $VERSION = '0.003'; ## VERSION
 ## AUTHORITY
 
 use Web::Async::WebSocket::Frame;
@@ -9,6 +9,7 @@ use Web::Async::WebSocket::Frame;
 use List::Util qw(pairmap);
 use Compress::Zlib;
 use POSIX ();
+use URI;
 use Time::Moment;
 use Digest::SHA qw(sha1);
 use MIME::Base64 qw(encode_base64);
@@ -33,6 +34,9 @@ our %COMPRESSIBLE_OPCODE = (
     $OPCODE_BY_NAME{binary} => 1,
 );
 
+# Whether we're `ws` or `wss`
+field $scheme : reader : param = 'ws';
+# The Web::Async::WebSocket::Server instance
 field $server : reader : param = undef;
 
 # Given the state of websockets in general, this is unlikely to change from `HTTP/1.1` anytime soon
@@ -53,6 +57,11 @@ field $supported_extension : reader : param {
     }
 }
 
+field $method : reader = undef;
+field $url : reader = undef;
+field $uri : reader = undef;
+field $headers : reader { +{ } }
+
 # What to report in the `Server:` header
 field $server_name : reader : param = 'perl';
 
@@ -198,7 +207,7 @@ method deflate ($data) {
     ) or die "Cannot create a deflation stream\n" ;
 
     my ($output, $status) = $deflation->deflate($data);
-    die "deflation failed\n" unless $status == Z_OK;
+    die "deflation failed - $status\n" unless $status == Z_OK;
     (my $block, $status) = $deflation->flush(Z_SYNC_FLUSH);
     die "deflation failed at flush stage\n" unless $status == Z_OK;
 
@@ -206,18 +215,17 @@ method deflate ($data) {
 }
 
 method inflate ($data) {
-    undef $inflation unless $compression_options->{cilent_context};
+    undef $inflation unless $compression_options->{client_context};
     $inflation //= inflateInit(
         -WindowBits => -($compression_options->{client_bits} || 15)
     ) or die "Cannot create a deflation stream\n" ;
 
     my ($block, $status) = $inflation->inflate($data);
-    die "deflation failed\n" unless $status == Z_STREAM_END or $status == Z_OK;
+    die "inflation failed - $status\n" unless $status == Z_STREAM_END or $status == Z_OK;
     return $block;
 }
 
 async method read_headers () {
-    my %hdr;
     while(1) {
         my $line = decode_utf8('' . await $stream->read_until("\x0D\x0A"));
         $line =~ s/\x0D\x0A$//;
@@ -225,9 +233,9 @@ async method read_headers () {
 
         my ($k, $v) = $line =~ /^([^:]+):\s+(.*)$/;
         $k = lc($k =~ tr{-}{_}r);
-        $hdr{$k} = $v;
+        $headers->{$k} = $v;
     }
-    return \%hdr;
+    return $headers;
 }
 
 method generate_response_key ($key) {
@@ -239,12 +247,21 @@ async method handle_connection () {
     try {
         $self->add_child($stream);
         my $first = await $stream->read_until("\x0D\x0A");
-        my ($method, $url, $version) = $first =~ m{^(\S+)\s+(\S+)\s+(HTTP/\d+\.\d+)\x0D\x0A$}a;
+        ($method, $url, my $version) = $first =~ m{^(\S+)\s+(\S+)\s+(HTTP/\d+\.\d+)\x0D\x0A$}a;
         $log->tracef('HTTP request is [%s] for [%s] version %s', $method, $url, $version);
         my $hdr = await $self->read_headers();
 
         $log->tracef('url = %s, headers = %s', $url, format_json_text($hdr));
 
+        # We rely on the caller to tell us the scheme, defaulting to plain `ws`,
+        # and everything else in the URI comes directly from the request.
+        $uri = URI->new($scheme . '://localhost');
+        $uri->host($hdr->{host}) if exists $hdr->{host};
+        $uri->path($url);
+
+        unless($hdr->{upgrade} =~ /^websocket$/i) {
+            die sprintf "No upgrade: websocket header, ignoring connection\n";
+        }
         unless($hdr->{sec_websocket_version} >= 13) {
             die sprintf "Invalid websocket version %s\n", $hdr->{sec_websocket_version};
         }
@@ -362,7 +379,7 @@ async method read_frame () {
         return await $self->close(
             code => 1002,
             reason => 'Unexpected reserved bit set',
-        ) if any { $_ } @rsv;
+        ) if any { $_ } @rsv[1..2];
         $type //= $opcode & 0x0F;
         return await $self->close(
             code   => 1002,
